

         /*********************************************************/
         /*                                                       */
         /*                 COPYRIGHT (C) 1988 BY                 */
         /* CAD Group of Dipartimento di Automatica e Informatica */
         /*             Politecnico di Torino, Italy.             */
         /*                  ALL RIGHTS RESERVED.                 */
         /*                                                       */
         /* This software is furnished under a license and may be */    
         /* used and copied  only in accordance with the terms of */
         /* such license. Unauthorized copies will be persecuted. */
         /*                                                       */
         /*********************************************************/





/******************************************************************************

    Filename:         topology.h

    Author:           Pier Luca Montessoro

    Creation date:    22 January 1988

    Contents:         data structures definitions used by the simulation kernel


                          *** DOCUMENTATION ***

      This file contains the definitions for the data structures representing
 the network to be simulated.

 IMPORTANT:  All the codes are defined and used as 'unsigned char'. This
             ensures the correct parameters passing between functions also
             on 16 bits word-width machines.


 SEMANTICS OF SIGNAL MODELING

      Creator is intended to be a multilevel simulator in the sense that 
 more signal models are allowed, each one giving a more accurate description 
 of the bewhavioural of the elements than the loweer level ones.

      The signal models currently implemented are:

 - Register Transfer Model

   It is used for boolean, RT and behavioural functions, and wires are the 
   carriers handling it. Its characteristics are:
   - static strength (independent for 0 and 1 values) for active elements
   - wires may have memory (the strength may derive from the devices 
     writing on them)
   - may be scalar or vectorial.

 - Switch Model

   It is used for the switch elements, i.e., transistors and nodes.
   Its characteristics are:
   - static strength for transistors
   - dynamic strength for nodes (maybe charged)
   - only scalar.

 - Electrical Logic model

   To be defined and implemented.


 ******************************************************************************/



/* codes related to the position of the element */

#define INTERNAL         (unsigned char)   0
#define INPUT            (unsigned char)   1
#define OUTPUT           (unsigned char)   2
#define BIDIRECTIONAL    (unsigned char)   3
#define PHANTOM          (unsigned char)   4
#define VDD              (unsigned char)   5
#define GND              (unsigned char)   6


/* codes related to the mlt type */

#define MLT_SINGLE_OUT   (unsigned char)   0
#define MLT_MULTI_OUT    (unsigned char)   1


/* codes related to the delay type */

#define ZERO_DELAY       (unsigned char)   0
#define SYMM_DELAY       (unsigned char)   1
#define ASYMM_DELAY      (unsigned char)   2
#define ZASYMM_DELAY     (unsigned char)   3
#define DYN_DELAY        (unsigned char)   4
#define ZDYN_DELAY       (unsigned char)   5
#define TRANSPORT_DELAY  (unsigned char)   6
#define NO_DELAY         (unsigned char)   7


/* codes related to the type of the topological connection */

#define NON_TRIGGERING   (unsigned char)   0
#define TRIGGERING       (unsigned char)   1
#define EVAL_FUNCTION    (unsigned char)   2


/* codes related to the type of value representation */

#define RT_MODEL         (unsigned char)   0
#define SWITCH_MODEL     (unsigned char)   1
#define E_LOGIC_MODEL    (unsigned char)   2
#define PTR_TO_FUNCTION  (unsigned char)   3


/* appendix types */

#define NOTHING              (unsigned char)   0  /* not really used as type */
#define FORCE                (unsigned char)   1
#define TRACE                (unsigned char)   2


/* modification codes */

#define UNCHANGED          (unsigned char)     0
#define EVALUATED          (unsigned char)     1
#define DISPOSED           (unsigned char)  0xFF


/* maximum cid. It is the (dummy) cid of the reference machine. */

#define MAX_CID    (unsigned short)  0xFFFF


/* carrier types */

#define NODE            (unsigned char)   0
#define WIRE            (unsigned char)   1
#define MEM_WIRE        (unsigned char)   2

#define FIRST_CARR_TYPE (unsigned char)   0
#define LAST_CARR_TYPE  (unsigned char)   2


/* constant type */

#define CONSTANT        (unsigned char)   3

#define FIRST_CONS_TYPE (unsigned char)   3
#define LAST_CONS_TYPE  (unsigned char)   3


/* boolean evaluation types */

#define AND             (unsigned char)   4
#define NAND            (unsigned char)   5
#define OR              (unsigned char)   6
#define NOR             (unsigned char)   7
#define BUF             (unsigned char)   8
#define INV             (unsigned char)   9
#define XOR             (unsigned char)  10
#define XNOR            (unsigned char)  11
#define TGP             (unsigned char)  12
#define TGN             (unsigned char)  13
#define MTGP            (unsigned char)  14
#define MTGN            (unsigned char)  15
#define BUS_TRI         (unsigned char)  16
#define BUS_MOS         (unsigned char)  17
#define BUS_UP          (unsigned char)  18
#define BUS_DOWN        (unsigned char)  19
#define SRNOR           (unsigned char)  20
#define SRNAND          (unsigned char)  21 
#define MUX             (unsigned char)  22 
#define AND_TRI         (unsigned char)  23 
#define NAND_TRI        (unsigned char)  24 
#define OR_TRI          (unsigned char)  25 
#define NOR_TRI         (unsigned char)  26 
#define BUF_TRI         (unsigned char)  27 
#define INV_TRI         (unsigned char)  28 
#define AND_TRIN        (unsigned char)  29 
#define NAND_TRIN       (unsigned char)  30 
#define OR_TRIN         (unsigned char)  31 
#define NOR_TRIN        (unsigned char)  32 
#define BUF_TRIN        (unsigned char)  33 
#define INV_TRIN        (unsigned char)  34 
#define AND_OC          (unsigned char)  35 
#define NAND_OC         (unsigned char)  36 
#define OR_OC           (unsigned char)  37 
#define NOR_OC          (unsigned char)  38 
#define BUF_OC          (unsigned char)  39 
#define INV_OC          (unsigned char)  40 
#define AND_OE          (unsigned char)  41 
#define NAND_OE         (unsigned char)  42 
#define OR_OE           (unsigned char)  43 
#define NOR_OE          (unsigned char)  44 
#define BUF_OE          (unsigned char)  45 
#define INV_OE          (unsigned char)  46

#define FIRST_GATE_TYPE (unsigned char)   4
#define LAST_GATE_TYPE  (unsigned char)  46


/* registers, flip-flops, etc. types */

#define DFF_PET         (unsigned char)  50
#define TFF_PET         (unsigned char)  51
#define JKFF_PET        (unsigned char)  52
#define DFF_PET_C       (unsigned char)  53
#define TFF_PET_C       (unsigned char)  54
#define JKFF_PET_C      (unsigned char)  55
#define DFF_PET_S       (unsigned char)  56
#define TFF_PET_S       (unsigned char)  57 
#define JKFF_PET_S      (unsigned char)  58 
#define DFF_PET_SC      (unsigned char)  59
#define TFF_PET_SC      (unsigned char)  60
#define JKFF_PET_SC     (unsigned char)  61
#define DFF_NET         (unsigned char)  62
#define TFF_NET         (unsigned char)  63
#define JKFF_NET        (unsigned char)  64
#define DFF_NET_C       (unsigned char)  65
#define TFF_NET_C       (unsigned char)  66
#define JKFF_NET_C      (unsigned char)  67
#define DFF_NET_S       (unsigned char)  68
#define TFF_NET_S       (unsigned char)  69
#define JKFF_NET_S      (unsigned char)  70
#define DFF_NET_SC      (unsigned char)  71
#define TFF_NET_SC      (unsigned char)  72
#define JKFF_NET_SC     (unsigned char)  73
#define LATCH_HIGH      (unsigned char)  74
#define LATCH_LOW       (unsigned char)  75

#define DFF_PET_CKS     (unsigned char)  76
#define DFF_PET_C_CKS   (unsigned char)  77
#define DFF_PET_S_CKS   (unsigned char)  78
#define DFF_PET_SC_CKS  (unsigned char)  79
#define DFF_NET_CKS     (unsigned char)  80
#define DFF_NET_C_CKS   (unsigned char)  81
#define DFF_NET_S_CKS   (unsigned char)  82
#define DFF_NET_SC_CKS  (unsigned char)  83

#define FIRST_REGI_TYPE (unsigned char)  50
#define LAST_REGI_TYPE  (unsigned char)  75
#define FIRST_CKS_TYPE  (unsigned char)  76
#define LAST_CKS_TYPE   (unsigned char)  83


/* switch level transistor types */

#define DEPL_TRANS      (unsigned char)  85 
#define N_CH_TRANS      (unsigned char)  86
#define P_CH_TRANS      (unsigned char)  87

#define FIRST_TRAN_TYPE (unsigned char)  85
#define LAST_TRAN_TYPE  (unsigned char)  87


/* functional element types */

#define REGISTER        (unsigned char)  90
#define ADD             (unsigned char)  91
#define ADC             (unsigned char)  92
#define SUB             (unsigned char)  93
#define SBB             (unsigned char)  94
#define MPY             (unsigned char)  95
#define DIV             (unsigned char)  96
#define ROR             (unsigned char)  97
#define ROL             (unsigned char)  98
#define SHL             (unsigned char)  99
#define SRA             (unsigned char) 100
#define SRL             (unsigned char) 101
#define INC             (unsigned char) 102
#define DEC             (unsigned char) 103
#define FAN             (unsigned char) 104
#define BIT_AND         (unsigned char) 105
#define BIT_OR          (unsigned char) 106
#define BIT_NOT         (unsigned char) 107
#define BIT_XOR         (unsigned char) 108
#define COMPR_AND       (unsigned char) 109
#define COMPR_OR        (unsigned char) 110
#define COMPR_XOR       (unsigned char) 111
#define CHS             (unsigned char) 112
#define IS_ZERO         (unsigned char) 113
#define IS_ONE          (unsigned char) 114
#define IS_DRIVEN       (unsigned char) 115
#define TO_CAR          (unsigned char) 116
#define CONCAT          (unsigned char) 117
#define EXTRACT         (unsigned char) 118
#define EQ              (unsigned char) 119
#define NE              (unsigned char) 120
#define GT              (unsigned char) 121
#define GE              (unsigned char) 122
#define LT              (unsigned char) 123
#define LE              (unsigned char) 124
#define TRIGGER_0       (unsigned char) 125
#define TRIGGER_1       (unsigned char) 126
#define PLA             (unsigned char) 127
#define RAM             (unsigned char) 128
#define ROM             (unsigned char) 129
#define COMPOUND        (unsigned char) 130

#define FIRST_FUNC_TYPE (unsigned char)  90
#define LAST_FUNC_TYPE  (unsigned char) 130


/* high speed evaluation gates */

#define AND_HS          (unsigned char) 140
#define NAND_HS         (unsigned char) 141
#define OR_HS           (unsigned char) 142
#define NOR_HS          (unsigned char) 143
#define BUF_HS          (unsigned char) 144
#define INV_HS          (unsigned char) 145
#define XOR_HS          (unsigned char) 146
#define XNOR_HS         (unsigned char) 147

#define FIRST_HS_TYPE   (unsigned char) 140
#define LAST_HS_TYPE    (unsigned char) 147


/* functional list identifier */

#define FUNCTION_LIST   (unsigned char) 255



/* machine dynamic descriptor definition:

   link                 is the pointer used to build the concurrent machines
                        The list is circular, the last concurrent machine 
                        points to the reference one.

   b_link               is the back pointer in concurrent machines list. 
                        Also the reverse list is circular, the reference 
                        machine points to the last concurrent one.

   actual_value         actual value of the element; for the functional 
                        lists, it contains the pointer to the evaluation 
                        function.

   next_value           future value, to become the actual one at the time
                        specified by the activation_time field. If the 
                        element is not scheduled, next value contains the 
                        INITIALIZATION_VALUE. This makes easier the 
                        convergence test of the concurrent machines.
                        If delay model is transport, the next value is used 
                        as pointer to the activity list, a list of 
                        machine descriptors (just to avoid memory 
                        fragmentation) containing the next activity 
                        information. 

   last_activation_time last time at which the element changed value. It
                        is used during the evaluation process to perform
                        the trigger inibition.

   next_activation_time time at which the element must be change value. It
                        contains MAX_UNSIGNED if the machine is quiescent.

   evaluated_value      is the word used to save the result of the
                        evaluation. It is used during the end_of_time function
                        to decide id the schedulation is needed or if the
                        concurrent machines must be explicitly represented.

   delay                is a structure where the delay information 
                        (static or dynamic) is stored at the beginning of 
                        the simulation and, if needed, after the evaluations.

   cid                  concurrent identifier.

   modify_flag          flag to signal what kind of modification has been
                        performed by the update or the evaluate function. */

typedef struct Machine
        {
           struct Machine        *link;

           struct Machine        *b_link;
                                 
           unsigned long         actual_value;
                                 
           unsigned long         next_value;
                                 
           unsigned long         last_activation_time;
                                 
           unsigned long         next_activation_time;
                                  
           unsigned long         evaluated_value;
                                  
           union Delay            
           {                      
              unsigned long      dynamic;
              unsigned long      symmetric;
              struct Asimmetrical
              {
                 unsigned short  rise;
                 unsigned short  fall;
              }  asymmetric;
           }  delay;

           unsigned short     cid;
           unsigned char      unused;
           unsigned char      modify_flag;

        }  MACHINE;



/* appendix definition:

   link                 is the pointer used to build the appendix list.
                        It contains the NULL value at the end of the
                        list.

   type                 code identifying the type of appendix.

   attr_1 ... attr_3    general purpose attributes.

   var_part             union containing different fields depending of the 
                        type of the appendix.

   value                (FORCE type) forced value.

   _0_strength, 
   _1_strength,         (FORCE type) forcing strength. 

   cid                  (TRACE type) Concurrent IDentifier of the 
                        concurrent machine to be traced. */

typedef struct Appendix
        {
           struct Appendix    *link;

           unsigned char      type;
           unsigned char      attr_1;
           unsigned char      attr_2;
           unsigned char      attr_3;

           union Var_Part
           {                                                       
              struct Force
              {
                 unsigned long      value;

                 unsigned char      _0_strength;
                 unsigned char      _1_strength;
                 unsigned char      unused_2;
                 unsigned char      unused_3;

              }  force;

              struct Trace
              {
                 unsigned long      cid;

                 unsigned long      unused;

              }  trace;

           } var_part;

        }  APPENDIX;



/* topological element definition:

   list_head            is the first pointer of the machine list. It always
                        points to the reference machine descriptor, which is
                        the first one.

   saved_list_head      is the first pointer of the saved machine list used 
                        during the restore operation.

   save_time            is set to the current time each time the machine 
                        list is saved.

   update_code          pointer to the function to be used to update the
                        element when active.

   mlt_code             pointer to the function to be used to perform the 
                        evaluation MLT for the element.

   postprocessing_code  pointer to the function to be used to postprocess the
                        element.

   type                 type of the element.

   driving_capabilities contains the information about the strength (or the
                        size, for nodes) of the elements. The strength is
                        coded using the least significant four bits of the
                        character for the zero value, and the other four for
                        the one value. The size is stored only in the least
                        significant half of the character.

   data_size            is the number of bits meaningful in the value fields.

   data_type            is a code indicating the value representation type.

   delay_type           is a code about the delay type: zero, symmetric, 
                        asimmetric (zero or non-zero) or dynamic.

   mlt_type             is a code about the mlt type: single output list or 
                        multiple output list. 

   position             is the code of the position of the element with
                        respect to the periphery on the network.

   detection_point      is a flag to signal that the element has been
                        selected to be a detection point.

   appendix             is the link to the list of miscellaneous information.

   persistence          is the time interval after machine update during 
                        which the timing information about the last transition 
                        time (last_update_time field) must be explicitely kept.

   level                levelizing value, i.e., index of the stack in which
                        the element has to be inserted to allow level-ordered
                        evaluations.

   fanin_position       is the address in the fanin vector at wich starts the
                        fanin information related to the topological element.

   fanout_position      is the address in the fanout vector at wich starts the
                        fanout information related to the topological element.

   fanin_number         is the number of fanin records related to the
                        topological element.

   fanout_number        is the number of fanout records related to the
                        topological element.

   eval_link            pointer used to link statically allocated eval link 
                        table for the element.

   stack_link           pointer used to link the topological descriptor in the
                        stack structure for levelized evaluation.

   last_update_time     is the simulation time of the last update of the
                        element list.

   cks_rise_phase       if the element is a clock input for an 
   cks_fall_phase       edge-sensitive element, or if it is on a clock 
                        distribution path, these fields contain the phase 
                        information for the edges of the clock. 

   cks_period           is the period of the clock to be suppressed. */

typedef struct Topological_Element
        {                                                      
           MACHINE                     *list_head;

           unsigned long               (*update_code)();

           int                         (*mlt_code)();

           int                         (*postprocessing_code)();

           unsigned                    type                 : 8;
           unsigned                    driving_capabilities : 8;
           unsigned                    data_type            : 3;
           unsigned                    data_size            : 5;
           unsigned                    mlt_type             : 1;
           unsigned                    delay_type           : 3;
           unsigned                    position             : 3;
           unsigned                    detection_point      : 1;

           APPENDIX                    *appendix;

           unsigned short              persistence;
           unsigned short              level;

           struct Fanin_Element        *fanin_position;

           struct Fanout_Element       *fanout_position;

           unsigned short              fanin_number;
           unsigned short              fanout_number;

           struct Eval_Link_Element    *eval_link;

           struct Topological_Element  *stack_link;

           unsigned long               last_update_time;

           unsigned short              cks_rise_phase;
           unsigned short              cks_fall_phase;

           unsigned short              cks_period;
           unsigned short              unused;

#ifdef RESTORE_ENABLED

           MACHINE                     *saved_list_head;

           unsigned long               save_time;

#endif

        }  TOPOLOGICAL_ELEMENT;



/* fanin element structure definition:

   link                 pointer to the fanin element.

   attribute            information about the relations between the element
                        itself and its fanin. */

typedef struct Fanin_Element
        {
           TOPOLOGICAL_ELEMENT  *link;

           unsigned char         attribute;
           unsigned char         unused1;
           unsigned short        unused2;

        }  FANIN_ELEMENT;



/* fanout element structure definition:

   link                 pointer to the fanout element.

   attribute            information about the relations between the element
                        itself and its fanout. */

typedef struct Fanout_Element
        {
           TOPOLOGICAL_ELEMENT  *link;

           unsigned char         attribute;
           unsigned char         unused1;
           unsigned short        unused2;

        }  FANOUT_ELEMENT;



/* evaluation link table: this is a table containing, for each element that 
   can be evaluated, the pre-allocated and pre-initialized data structure  
   needed to perform the E-MLT. The element definition is the following:

   curr_pointer         pointer to the current machine descriptor.

   ref_pointer          pointer to the reference machine descriptor.

   obl_pointer          obligation pointer.

   mask                 word used to recognize the bits that must be read
                        from the data field of the element to perform the
                        evaluation.

   attribute            information about the relations between the element
                        itself and the one to be read. 

   delay_type           is the type of delay. It is used because when an 
                        element is disposed, if its delay model is transport, 
                        also the activity list must be processed. */

typedef struct Eval_Link_Element
        {
           MACHINE         *curr_pointer;

           MACHINE         *ref_pointer;

           MACHINE         *obl_pointer;

           unsigned long   mask;

           unsigned char   attribute;
           unsigned char   delay_type;
           unsigned short  unused;

        }  EVAL_LINK_ELEMENT;

